# Python3 script to test the algorithm of the MTC3 challenge:
#   A CLOAKED SUBSTITUTION CIPHER – PART 1
# Script shows the encryption and decryption process
# Author of this script: GT (modifications by BE)
# August 2019

import sys
import os
import random
import argparse

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

version = "v1.2 (2019-08-09)" # added compared to v1.1.2: options -kgs, -w, -o

# define alphabet
A_S = "ABCDEFGHIJKLMNOPQRSTUVWXYZ._,-?"  # plaintext alphabet (from v1.1.2: _. ==> ._)
A_N = "abcdefghijklmnopqrstuvwxyz*^’=+"  # additional character set (assure that none of them is in A_S)
A = A_S + A_N                            # ciphertext alphabet


def create_random_key(size):
    return ''.join(random.choice(A) for i in range(size))


"""This function does the simple encryption process, using as arguments the plaintext and the key"""
def simple_encryption(plaintext,key):
        # Simple Substitution Encryption Algorithm C <== E(K,M)
        # Replace each character m of M by jK(m).
        # and a ciphertext cryptogram C is decrypted into a plaintext message
        # M using the key K by means of the following
        
        ciphertext = ""
        for i in range(0,len(plaintext)):
                for j in range(0,len(A_S)):
                        if plaintext[i:i+1] == A_S[j:j+1]:
                                index_plaintext_letter = j
                                break
                ciphertext_char = ""
                ciphertext_char = key[index_plaintext_letter:index_plaintext_letter+1]
                ciphertext += ciphertext_char
                
        return(ciphertext)

"""This function does the simple decryption process, using as arguments the ciphertext and the key"""
def simple_decryption(ciphertext,key):
        # Simple Substitution Decryption algorithm: M <== D(K,C)
        # Replace each character c of C by jK^-1(c).

        plaintext = ""
        for i in range(0,len(ciphertext)):
                for j in range(0,len(A)):
                        if ciphertext[i:i+1] == key[j:j+1]:
                                index_ciphertext_letter = j
                                break
                plaintext_char = ""
                plaintext_char = A_S[j:j+1]
                plaintext += plaintext_char
                
        return(plaintext)

"""This function produces the Signal key, according to the key"""
def create_Signal_key(key, A):
        S_K = []

        for i in range(0,len(A)):
                if A[i:i+1] in key:
                        S_K.append(A[i:i+1])
        SK = ""
        for i in range(0,len(S_K)):
                SK += S_K[i]
        
        return(SK)

"""This function produces the Noise key, according to the key"""        
def create_Noise_key(key, A):
        N_K = []

        for i in range(0,len(A)):
                if A[i:i+1] not in key:
                        N_K.append(A[i:i+1])

        NK = ""
        for i in range(0,len(N_K)):
                NK += N_K[i]

        return(NK)

"""This function does the cloaked substitution encryption process, using as arguments the plaintext and the key (noise characters randomly inserted)"""
def cloaked_substitution_encryption(plaintext,key,NK):
        ciphertext = simple_encryption(plaintext,key)
        i = -1  # BE: changed from 0 to -1 as index in Python starts with 0
        j = -1
        C_ = []
        while j <= len(ciphertext):
                # flip a coin
                if random.random() <= 0.49:
                        coin_condition = "heads"
                else:
                        coin_condition = "tails"
                #print(coin_condition)
                if coin_condition == "heads":
                        choice = int(random.random()*len(NK))
                        #print("choice=",choice)
                        noise_character = NK[choice:choice+1]
                        i += 1
                        C_.append(noise_character)
                else:
                        j += 1 # BE: paper was right, when initial index set ok
                        if j <= len(ciphertext):
                                i += 1
                                C_.append(ciphertext[j:j+1])
        C = ""
        for i in range(0,len(C_)):
                C += C_[i]
        return(C)

"""This function does the cloaked substitution decryption process, using as arguments the ciphertext and the key"""
def cloaked_substitution_decryption(key,C,NK):
        ciphertext = ""
        for i in range(0,len(C)):
                if C[i:i+1] not in NK:
                        ciphertext += C[i:i+1]
        # Now ciphertext contains the content after intermediate step 1
        plaintext = simple_decryption(ciphertext,key)
        return(plaintext, ciphertext)

"""This function generates a random key. It returns a key with length 62 symbols, but we use as the key, the first 31 symbols"""
def generate_random_key(total_alphabet):
        A_ = []
        for i in range(0,len(total_alphabet)):
                A_.append(total_alphabet[i:i+1])
        for i in range(0,len(total_alphabet)):
                place_1 = randint(0,len(total_alphabet)-1)
                place_2 = randint(0,len(total_alphabet)-1)
                while place_1 == place_2:
                        place_2 = randint(0,len(total_alphabet)-1)
                swap = A_[place_1]
                A_[place_1] = A_[place_2]
                A_[place_2] = swap
        A_random = ""
        for i in range(0,len(total_alphabet)):
                A_random += A_[i]
        random_key = A_random[0:31]
        #print(random_key + "\n")
        return(A_random)

"""This function keeps all capital letters from the full plaintext. This enables us to score this plaintext with 4grams"""
def simplify_plaintext(full_plaintext):
        only_capitals_plaintext = ""
        for i in range(0,len(full_plaintext)):
                if (ord(full_plaintext[i]) >= 65 and ord(full_plaintext[i]) <= 90):
                        only_capitals_plaintext += full_plaintext[i]
        return(only_capitals_plaintext)


''' writes on stderr '''
def printinfo():  # used by test1() and when option -v
    # eprint("\n" + printinfo.__name__)
    eprint()
    # eprint("Plaintext alphabet  (length " + str(len(A_S)) + "): " + A_S)
    # eprint("Ciphertext alphabet (length " + str(len(A)) + "): " + A)
    eprint("Key (length " + str(len(key)) + "): " + key)


''' Use test data from the example in the challenge template (if nothing else is given at the commandline)'''
def test1():
    # Declaration "global" only needed if you like to write to a global variable
    # [here, this is necessary only if the global variable is used within a function called by test1()]
    # As test1() is called several times don't change args components within this function.
    # global ciphertext, plaintext, key
    global key

    key = "Dx-UM^QpKOs._YfBzJrLhAaTCgbNmGW"
    PT = "THE_QUICK,_BROWN_FOX_JUMPS._OVER_A_LAZY-DOG?"

    print("\nTest data:")
    print('PT ALPHABET (length ' + str(len(A_S)) + '): ' + A_S)
    print('CT ALPHABET (length ' + str(len(A)) + '): ' + A)
    print('KEY (length ' + str(len(key)) + '): ' + key)
    print('PT (length ' + str(len(PT)) + '): ' + PT)
    print()

    # Expected ciphertext from simple encryption (output always the same for given key and plaintext):
    # "LpMbzhK-smbxJfaYb^fTbOh_BrNbfAMJbDb.DgCGUfQW"
    print("SIMPLE ENCRYPTION RESULT: Plaintext is encrypted to:")
    CT1 = simple_encryption(PT,key)
    print(CT1)
    print()

    PT1_dec = simple_decryption(CT1,key)
    print("SIMPLE DECRYPTION RESULT: Ciphertext (length " + str(len(CT1)) + ") is decrypted to:")
    print(PT1_dec)
    print()

    # Expected result:
    # SIGNAL KEY= ABCDGJKLMNOQTUWY_.-abfghmprsxz^
    # NOISE KEY = EFHIPRSVXZ,?cdeijklnoqtuvwy*’=+
    SK = create_Signal_key(key,A)
    NK = create_Noise_key(key,A)
    print("SIGNAL KEY=",SK)
    # print()
    print("NOISE KEY =",NK)
    print()

    # Possible expected ciphertext from cloaked substitution encryption (random output):
    # CT2="LqcjpMbFztedhyvK-smbxiJfaVX=YVSqb^w,fTibkOh’_BrNFPbo*eZ+fnZAcMJjbDIwb=.DugC’wj=tG’kqZUf?QkW"
    CT2 = cloaked_substitution_encryption(PT,key,NK)
    print("CLOAKED SUBSTITUTION ENCRYPTION: Ciphertext (length " + str(len(CT2)) + "):")
    print(CT2)
    print()
    # if args.verbose: printinfo()

    PT2_dec, strippedCT = cloaked_substitution_decryption(key,CT2,NK)
    print("CLOAKED SUBSTITUTION DECRYPTION: 1. Ciphertext without noise chars (length " + str(len(strippedCT)) + "); 2. Found plaintext")
    print(strippedCT)
    print(PT2_dec)
    print()
    # if args.verbose: printinfo()

    print("SIMPLIFYING PLAINTEXT:")
    print(PT2_dec)
    print("becomes...")
    print(simplify_plaintext(PT2_dec))
    print()




######### main() #################
if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="output additional information on stderr", action="count", default=0)

    parser.add_argument("-w", "--writefinalresulttofile", help="if input string is read from file, the final result (plaintext or ciphertext) is written in addition to file 'result_*.txt'", action="count", default=0)

    mgroup1 = parser.add_mutually_exclusive_group()
    mgroup1.add_argument("-kgs", "--generaterandomkeystring", help="generate random key string and write it to commandline", action="count", default=0)

    mgroup2 = parser.add_mutually_exclusive_group()
    mgroup2.add_argument("-krs", "--readkeystring", metavar="STRING", help="read STRING from commandline as key")
    mgroup2.add_argument("-krf", "--readkeyfile", metavar="FILE", help="read key from FILE")

    mgroup4 = parser.add_mutually_exclusive_group()
    mgroup4.add_argument("-es", "--encryptstring", metavar="STRING", help="encrypt STRING from commandline")
    mgroup4.add_argument("-ef", "--encryptfile", metavar="FILE", help="read plaintext from FILE and encrypt it")
    mgroup4.add_argument("-ds", "--decryptstring", metavar="STRING", help="decrypt STRING from commandline")
    mgroup4.add_argument("-df", "--decryptfile", metavar="FILE", help="read ciphertext from FILE and decrypt it")
    mgroup4.add_argument("-t", "--test", help="encrypt and decrypt a string with a given key (using sample data from the MTC3 template)", action="store_true")

    args = parser.parse_args()

    if len(sys.argv)==1:
        print('\nPython3 script to show the encryption and decryption process of the MTC3 challenge "A CLOAKED SUBSTITUTION CIPHER – PART 1"\n')
        print('Typical calls: python program.py')
        print('               python program.py -t -v')
        print('               python program.py -kgs -v')
        print('               python program.py -krs "Dx-UM^QpKOs._YfBzJrLhAaTCgbNmGW" -es "THE_QUICK,_BROWN_FOX_JUMPS._OVER_A_LAZY-DOG?"')
        print('               python program.py -krs "Dx-UM^QpKOs._YfBzJrLhAaTCgbNmGW" -ds "=ILHEcZcpq*MbyXzwhRoK-+sHmbxZJfFaYdb=k^tk,*SfTbOh?d_BrNbqfA?MJubDbv.DRgCGdURZvyfXQWZ"\n')
        print('               python program.py -krf key1.txt -ef plaintext1.txt -w')
        parser.print_help()
        sys.exit(1)
    if args.verbose:
        print('Version: ' + version)
        if len(sys.argv)==2:
            # parser.print_help()
            sys.exit(1)

    # generate key
    if args.generaterandomkeystring:
        key = create_random_key(len(A_S))
        SK = create_Signal_key(key,A)
        NK = create_Noise_key(key,A)
        print('KEY (length ' + str(len(key)) + '): ' + key)
        print("SIGNAL KEY=",SK)
        print("NOISE KEY =",NK)
        print()
        sys.exit(1)

    # read key
    if args.readkeyfile:
        key = open(args.readkeyfile, 'r').read().rstrip('\r\n')
    if args.readkeystring:
        key = args.readkeystring;
    if args.readkeyfile or args.readkeystring:
        SK = create_Signal_key(key,A)
        NK = create_Noise_key(key,A)	

    # encrypt / decrypt / test

    if args.encryptfile:
        args.encryptstring = open(args.encryptfile, 'r').read().rstrip('\r\n')

    if args.decryptfile:
        args.decryptstring = open(args.decryptfile, 'r').read().rstrip('\r\n')

    if args.encryptstring:
        PT = args.encryptstring
        CT = cloaked_substitution_encryption(PT,key,NK)
        if args.verbose:
            printinfo()
            print("\nPlaintext (length " + str(len(PT)) + "): " + PT)
        print("\nCiphertext (length " + str(len(CT)) + "): " + CT)
        print()
        if args.encryptfile and args.writefinalresulttofile:
            outputfilename = "result_" + args.encryptfile
            if args.readkeyfile:
                outputfilename = outputfilename + "_" + args.readkeyfile
            # print("Argument -w. Name = " + outputfilename)
            open(outputfilename, 'w').write(CT)

    elif args.decryptstring:
        CT = args.decryptstring
        PT,strippedCT = cloaked_substitution_decryption(key,CT,NK)
        if args.verbose:
            printinfo()
            print("\nCiphertext (length " + str(len(CT)) + "): " + CT)
            print("\nstrippedCT (length " + str(len(strippedCT)) + "): " + strippedCT)
        print("\nPlaintext  (length " + str(len(PT)) + "): " + PT)
        print()
        if args.decryptfile and args.writefinalresulttofile:
            outputfilename = "result_" + args.decryptfile
            if args.readkeyfile:
                outputfilename = outputfilename + "_" + args.readkeyfile
            # print("Argument -w. Name = " + outputfilename)
            open(outputfilename, 'w').write(PT)

    elif args.test:
        test1()
        sys.exit(1)

    else: # you'll never get here (handled via argparse)
        # if args.verbose: print('Version   : ' + version)
        print('PT ALPHABET (length ' + str(len(A_S)) + '): ' + A_S)
        print('CT ALPHABET (length ' + str(len(A)) + '): ' + A)
        print('KEY (length ' + str(len(key)) + '): ' + key)


